--[[
This file is part of Exploding numbers - a puzzle game

Exploding numbers is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
]]


-- replaces functions from the main game
-- old functions are saved as rr.old_xx
function hijack(mini)
	local h = mini.hijacked_love_functions
	for a = 1, #h do
		mini['old_' .. h[a]] = love[h[a]]
		love[h[a]] = mini[h[a]]
	end
	local h = mini.hijacked_globals
	if not h then return end
	for a = 1, #h do
		mini['old_' .. h[a]] = _G[h[a]]
		_G[h[a]] = mini[h[a]]
	end
end


-- restore hijacked functions and globals
function release(mini)
	local h = mini.hijacked_love_functions
	for a = 1, #h do
		love[h[a]] = mini['old_' .. h[a]]
	end
	local h = mini.hijacked_globals
	if not h then return end
	for a = 1, #h do
		_G[h[a]] = mini['old_' .. h[a]]
	end
end


help = {

text = function(self)

	local t = {[[*Exploding numbers *v0.96* - a puzzle game
Copyright © 2022, 2023 Pajo (xpio]] .. string.char(64) .. [[tut.by)


RULES*

• Even numbers can be divided by clicking at empty tile *next to* them. (One half gets blown away from the cursor.)*
]]
, -- figure 1
[[**
• Numbers divisible by four can also be exploded in four directions by clicking *on* them.*
]]
, -- figure 2
[[**
• Odd numbers can't be moved, but all numbers can combine with blown parts of other numbers.*
]]
, -- figure 3
[[**
• When the stated goal is reached, click *on* that number to progress to the next level.


*UNLOCKABLES*

Solve ... levels - to unlock ...
*25%* - all levels
*50%* - editor
*75%* - shapes
*all* - a few color schemes


*KEYBOARD SHORTCUTS*

Working everywhere:
*Esc* - exit submenu or quit game
*q* - quit game immediately
(*Shift* +) *c* - change color scheme

In help:
*F1* - exit help
*arrows*, *hjkl*, *PageUp*, *PageDown*, *Home*, *End*, *Space* and *b* - scroll the text

When playfield is shown:
*arrows* / *hjkl* - select tile
*Enter* / *Space* - click on selected tile

Other shortcuts are hinted on menu buttons by big letters or capital letters in their titles.


*MOUSE / TOUCH SHORTCUTS*

At the playfield:
*touch* / *click* - select
*drag* / (*Shift* +) *wheel* - move playfield if the playfield is wider than window
*Ctrl* + *wheel* / *pinch* gesture - set minimum zoom level (see *Settings*)

At the menu:
*touch* / *click* - select
*drag* / *wheel* - move menu left - right if the menu is wider than window
*Ctrl* + *wheel* - resize menu buttons


*MAIN MENU*

*Goal* - toggle the menu (Current level's goal is shown.)

*Level -* / *Restart* (F5) - restart level if the level is changed, or skip to previous level
*Level +* / *Next* (F6) - skip to next level

*Undo* and *Redo* / *Solve* - review all steps of either the built-in solution, or your own actions
Restart level to reload the built-in solution.


*SETTINGS*

*Editing* - toggle *Edit* button and help section
*Fullscr.* - toggle fullscreen / windowed mode
*Buttons* - change menu size between 60 (small) and 100 (big)
*Color scheme*, *Shape* - change the visual appearance

Current state, settings, custom levels and list of solved levels are saved in a file called *save*. The last exported level is saved in a readable file called *exported_level*. These files are in *]] .. love.filesystem.getSaveDirectory() .. [[* directory.
*Reset* - deletes them and exits the game

*Minimal zoom*
The entire level is visible by default, which may result in tiny numbers. This option sets a limit to zooming-out. It can also be set with pinching gesture, or *Ctrl* + *wheel* at the playfield. The minimum circle size, if smaller than default, will be shown in the top-left circle. In case some parts of the level end up off screen, use arrows, (*Shift* +) wheel, or dragging to pan the view.

]] .. (settings.editing and
[[

*EDITING*

Click on *Edit* button in main menu to start editing.
When editing a built-in level, it will first be copied as a new custom level. Custom levels are marked with *º*.

While in Edit mode, clicking on the grid places new numbers. Click a number with the same value to remove it.
*Value* - choose a value
*Goal* - set the goal for the level
*Step* - for *Value* and *Goal* buttons

*Copy* - save the current state as a new level
*Save* - save as the current level
These buttons exit the editing mode.

*Clear* - empty the level
*Rewind* - undo all changes
Both actions can be reversed by *Undo* and *Redo*.
Saving a cleared level will delete it. Saving a rewound level will make no changes and keep the edited level's solution, if any.

*Export* - dump the current state of the edited level to clipboard, standard output and a file in Love's directory, without saving it among levels
See *Settings* section for the directory's path.
*Import* - import a level from clipboard as a new level
The level in clipboard should be delimited by *{* and *}*.

]] or '') ..
[[

*LICENSE*

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see <*http://www.gnu.org/licenses/*>.*
]]}

	local colored_text = {}
	for fragment = 1, #t do
		colored_text[fragment] = {}
		for match in t[fragment]:gmatch('(.-)%*') do
			table.insert(colored_text[fragment], match)
			table.insert(colored_text[fragment], (#colored_text[fragment] % 4 == 1)
				and self.heading_c or self.text_c
			)
		end
	end
	return colored_text
end,


hijacked_love_functions = { 'draw', 'resize', 'update', 'mousepressed',
	'mousemoved', 'mousereleased', 'wheelmoved', 'keypressed'
},


draw = function() -- hijacked function
	local fh = help.font_height
	function circle(number, color, x, y)
		set_color(color)
		if settings.rectangles == 1 then
			love.graphics.rectangle('fill', x - fh * .2, y + fh * .05, fh * .9, fh * .9)
		else
			love.graphics.circle('fill', x + fh / 4, y + fh / 2, fh / 2)
		end
		set_color(color .. '_numbers')
		if number == '@' then x, y = x - fh * .16, y - fh * .1 end
		love.graphics.print(number, x, y)
	end
	function arrow(x, y)
		set_color('help_text')
		love.graphics.setLineWidth(2)
		love.graphics.line(x, y + fh / 2, x + fh / 2, y + fh / 2)
		love.graphics.line(
			x + fh / 4, y + fh / 4,
			x + fh / 2 , y + fh / 2,
			x + fh / 4, y + fh * 3 / 4
		)
	end
	love.graphics.setFont(help.font)
	love.graphics.clear(help.bg_c)
	love.graphics.setColor(1,1,1, 1)
	local h = help.y + help.top_margin
	love.graphics.draw(help.formatted, help.x, h)
	-- fig 1
	circle('', 'selected', help.x + fh * 2, h + help.height1 + fh * 1.5)
	circle('2', 'db2', help.x + fh * 3, h + help.height1 + fh * 1.5)
	arrow(help.x + fh * 4.5, h + help.height1 + fh * 1.5)
	circle('@', 'selected', help.x + fh * 6, h + help.height1 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 7, h + help.height1 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 8, h + help.height1 + fh * 1.5)
	-- fig 2
	circle('4', 'selected', help.x + fh * 2, h + help.height2 + fh * 1.5)
	arrow(help.x + fh * 3.5, h + help.height2 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 5, h + help.height2 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 6, h + help.height2 + fh * .5)
	circle('@', 'selected', help.x + fh * 6, h + help.height2 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 6, h + help.height2 + fh * 2.5)
	circle('1', 'odd', help.x + fh * 7, h + help.height2 + fh * 1.5)
	-- fig 3
	circle('3', 'odd', help.x + fh * 2, h + help.height3 + fh * 1.5)
	circle('', 'selected', help.x + fh * 3, h + help.height3 + fh * 1.5)
	circle('2', 'db2', help.x + fh * 4, h + help.height3 + fh * 1.5)
	circle('3', 'odd', help.x + fh * 5, h + help.height3 + fh * 1.5)
	arrow(help.x + fh * 6.5, h + help.height3 + fh * 1.5)
	circle('3', 'odd', help.x + fh * 8, h + help.height3 + fh * 1.5)
	circle('@', 'selected', help.x + fh * 9, h + help.height3 + fh * 1.5)
	circle('1', 'odd', help.x + fh * 10, h + help.height3 + fh * 1.5)
	circle('4', 'db4', help.x + fh * 11, h + help.height3 + fh * 1.5)
	help:draw_status()
end,


move = function(self, dy)
	screen.changed = true
	local oldy = self.y
	self.y = self.y + dy
	self.y = math.min(0, self.y)
	self.y = math.max(self.miny, self.y)
end,


keypressed = function(k) -- hijacked function
	if k == 'escape' or k == 'f1' then
		help:quit()
	elseif k == 'q' then
		love.event.push('quit')
	elseif k == 'pagedown' or k == 'l' or k == 'space' then
		help:move(help.top_margin + help.font_height - screen.h)
	elseif k == 'pageup' or k == 'h' or k == 'b' then
		help:move(screen.h - help.font_height - help.top_margin)
	elseif k == 'end' then
		help:move(- help.height)
	elseif k == 'home' then
		help:move(help.height)
	elseif k == 'c' then
		local lk_id = love.keyboard.isDown
		menu['colors_' .. (
			(lk_id('lshift') or lk_id('rshift')) and 'prev' or 'next'
		)]:onclick()
		help:init_colors()
		help.formatted = love.graphics.newText(help.font, '')
		local txt = help:text()
		local h = 0
		for fragment = 1, 4 do
			help.formatted:addf(txt[fragment], help.w, 'left', nil, h)
			local h0 = help.formatted:getHeight() + h
			h = h0 + 3 * help.font_height
		end
	end
end,


wheelmoved = function(x, y) -- hijacked function
	help:move(help.font_height * y * 2)
end,


mousepressed = function(mx, my, button) -- hijacked function
	mx, my = love.graphics.inverseTransformPoint(mx, my)
	if my < settings.menu_size then
		if mx >= help.button_x then
			help:quit()
		else
			help.scroll = true
			help.y = help.miny * mx / help.button_x
			help:move(0)
		end
	elseif button == 1 then
		help.drag = true
	end
end,


mousereleased = function(mx, my, button) -- hijacked function
	help.drag, help.scroll = false, false
	help.speed = help.speed * 2
end,


mousemoved = function(x, y, dx, dy) -- hijacked function
	x, y = love.graphics.inverseTransformPoint(x, y)
	if help.drag then
		help.speed = dy
		help:move(dy)
	end
	if help.scroll then
		help.y = help.miny * x / help.button_x
		help:move(0)
	end
end,


draw_status = function(self)
	local h = settings.menu_size
	local v = (self.miny == 0) and 'F1' or math.floor(self.y / self.miny * 100 + .5) .. '%'
	if v ~= 'F1' then
		-- progress bar
		if settings.menu_size > screen.h / 8 then h = h / 2 end
		love.graphics.setColor(self.progress_bg_c)
		love.graphics.rectangle('fill',
			0, 0,
			screen.w, h
		)
		love.graphics.setColor(self.progress_c)
		love.graphics.rectangle('fill',
			0, 0,
			self.button_x * self.y / self.miny, h
		)
	end
	-- button
	love.graphics.setColor(self.button_bg_c)
	love.graphics.rectangle('fill',
		self.button_x, 0,
		screen.w - self.button_x, h
	)
	love.graphics.setColor(self.button_text_c)
	love.graphics.setFont(menu_font_small)
	love.graphics.print('Return', self.button_text_x, self.button_text_y)
	love.graphics.setFont(self.font)
	if h == settings.menu_size then
		love.graphics.printf(v, self.button_x, self.button_text2_y, settings.menu_size, 'center')
	elseif v ~= 'F1' then
		love.graphics.print(v, 4, 2)
	end
end,


init_colors = function(self)
	local c = colors[settings.color_scheme]
	self.bg_c = c.help_bg or c.bg
	self.text_c = c.help_text or c.menu_text
	self.heading_c = c.heading
	self.progress_bg_c = {unpack(c.menu_bg)}
	self.progress_bg_c[4] = .6 -- add transparency
	self.progress_c = c.progress
	self.button_bg_c = c.menu_bg
	self.button_text_c = c.menu_text
end,


init = function(self)
	screen.changed = true
	local FONT_SIZE = 20
	self.w = screen.w
	if self.w > 450 then
		self.x = 25
		self.w = self.w - 2 * self.x
	else
		self.x = 0
	end
	self:init_colors()
	self.font = love.graphics.newFont(FONT_SIZE)
	self.font_height = self.font:getHeight()
	self.formatted = love.graphics.newText(self.font, '')
	local txt = self:text()
	local h = 0
	for fragment = 1, 4 do
		self.formatted:addf(txt[fragment], help.w, 'left', nil, h)
		local h0 = self.formatted:getHeight() + h
		self['height' .. (fragment < 4 and fragment or '')] = h0
		h = h0 + 3 * self.font_height
	end
	self.button_x = screen.w - settings.menu_size
	self.top_margin = settings.menu_size
	self.miny = math.min(0, screen.h - self.height - 2 * self.top_margin)
	self.y = math.max(self.y or 0, self.miny)
	self.drag = false
	self.scroll = false
	love.keyboard.setKeyRepeat(true)
	help.speed = 0
	self.button_text_x = self.button_x + (settings.menu_size - menu_font_small:getWidth('Return')) / 2 + .5
	self.button_text_y = settings.menu_size * .1
	self.button_text2_y = settings.menu_size / 2
end,


main = function(self)
	hijack(self)
	self:init()
end,


update = function(dt) -- hijacked function
	if dt < 1/50 then
		love.timer.sleep(1/30 - dt)
	end
	if
		love.keyboard.isDown('down') or love.keyboard.isDown('j') or
		love.mouse.isDown(2)
	then
		help:move(- help.font_height / 10)
	end
	if
		love.keyboard.isDown('up') or love.keyboard.isDown('k') or
		love.mouse.isDown(3)
	then
		help:move(help.font_height / 10)
	end
	if not help.drag and help.speed ~= 0 then
		help:move(help.speed)
		help.speed = help.speed / 1.1
		if math.abs(help.speed) < .1 then help.speed = 0 end
	end
	check_rotation(dt)
end,


resize = function() -- hijacked function
	love.graphics.origin()
	local x, y, w, h = love.window.getSafeArea()
	-- prevent crash on very small window
	if w < 2 * settings.menu_size or h < 2 * settings.menu_size then return end
	screen.w, screen.h = w, h
	love.graphics.translate(x, y)
	resize_grid()
	help:init()
end,


quit = function(self)
	release(self)
	love.resize()
	love.keyboard.setKeyRepeat(false)
end,


} -- end of help
